/*
 * Copyright (c) Andrey Kuznetsov. All Rights Reserved.
 *
 * http://uio.imagero.com
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *  o Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *
 *  o Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *
 *  o Neither the name of Andrey Kuznetsov nor the names of
 *    its contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
 * OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

package com.imagero.uio;

import com.imagero.uio.buffer.Buffer;
import com.imagero.uio.buffer.BufferManager;
import com.imagero.uio.buffer.MutableBufferManager;
import com.imagero.uio.buffer.MutableRAFBufferManager;
import com.imagero.uio.buffer.RABufferManager;
import com.imagero.uio.buffer.RAFBufferManager;
import com.imagero.uio.buffer.MutableRABufferManager;
import com.imagero.uio.buffer.HTTPBufferManager;
import com.imagero.uio.buffer.InputStreamBufferManager;
import com.imagero.uio.buffer.OutputStreamBufferManager;
import com.imagero.uio.buffer.arrays.ShortArrayBufferManager;
import com.imagero.uio.buffer.arrays.CharArrayBufferManager;
import com.imagero.uio.buffer.arrays.IntArrayBufferManager;
import com.imagero.uio.buffer.arrays.LongArrayBufferManager;
import com.imagero.uio.buffer.arrays.FloatArrayBufferManager;
import com.imagero.uio.buffer.arrays.DoubleArrayBufferManager;

import java.io.File;
import java.io.IOException;
import java.io.RandomAccessFile;
import java.io.InputStream;
import java.io.OutputStream;
import java.net.URL;

/**
 * UIOStreamBuilder is a builder pattern implementation and replacement for RandomAccessFactory.
 * Usual process looks like
 * File f = ...;
 * RandomAccessRO ro = new UIOStreamBuilder(f).setByteOrder(RandomAccess.LITTLE_ENDIAN).setBuffered(true).create();
 * or
 * RandomAccess ra = (RandomAccess)new UIOStreamBuilder(f).setMode(UIOStreamBuilder.READ_WRITE).create();
 *
 * Defaul values are:
 * mode - UIOStreamBuilder.READ_ONLY
 * byte order - RandomAccess.BIG_ENDIAN
 * buffered - false (however some streams are always buffered)
 *
 * @see #create
 * @see #setBuffered
 * @see #setByteOrder
 * @see #setMode
 * @see #setStart
 * @see #setLength
 *
 * @author Andrey Kuznetsov
 */
public class UIOStreamBuilder {

    public static final String READ_ONLY = "r";
    public static final String READ_WRITE = "rw";

    String mode = READ_ONLY;

    int byteOrder;

    Long start;
    Long length;

    File fileSource;

    RandomAccessFile rafSource;

    RandomAccess raSource;
    RandomAccessRO roSource;

    byte[] byteSource;
    short[] shortSource;
    char[] charSource;
    int[] intSource;
    long[] longSource;
    float[] floatSource;
    double[] doubleSource;

    Buffer[] bufferSource;

    InputStream inputStreamSource;
    OutputStream outputStreamSource;

    BufferManager bufferManager;
    MutableBufferManager mutableBufferManager;

    URL url;

    boolean buffered;

    Creator creator;

    public UIOStreamBuilder(String filename) {
        this(new File(filename));
    }

    public UIOStreamBuilder(File file) {
        this.fileSource = file;
        this.creator = new FileCreator();
    }

    public UIOStreamBuilder(RandomAccessFile rafSource) {
        this.rafSource = rafSource;
        this.creator = new RAFCreator();
    }

    /**
     * always buffered
     * @param ra
     */
    public UIOStreamBuilder(RandomAccess ra) {
        this.raSource = ra;
        this.creator = new RACreator();
    }

    /**
     * always buffered
     * @param ro
     */
    public UIOStreamBuilder(RandomAccessRO ro) {
        this.roSource = ro;
        this.creator = new ROCreator();
    }

    public UIOStreamBuilder(byte[] byteSource) {
        this.byteSource = byteSource;
        this.creator = new ByteCreator();
    }

    /**
     * always buffered
     * @param shortSource
     */
    public UIOStreamBuilder(short[] shortSource) {
        this.shortSource = shortSource;
        creator = new ShortCreator();
    }

    /**
     * always buffered
     * @param intSource
     */
    public UIOStreamBuilder(int[] intSource) {
        this.intSource = intSource;
        creator = new IntCreator();
    }

    /**
     * always buffered
     * @param longSource
     */
    public UIOStreamBuilder(long[] longSource) {
        this.longSource = longSource;
        creator = new LongCreator();
    }

    /**
     * always buffered
     * @param floatSource
     */
    public UIOStreamBuilder(float[] floatSource) {
        this.floatSource = floatSource;
        creator = new FloatCreator();
    }

    /**
     * always buffered
     * @param doubleSource
     */
    public UIOStreamBuilder(double[] doubleSource) {
        this.doubleSource = doubleSource;
        creator = new DoubleCreator();
    }

    /**
     * always buffered
     * @param manager
     */
    public UIOStreamBuilder(BufferManager manager) {
        this.bufferManager = manager;
        creator = new BMCreator();
    }

    /**
     * always buffered
     * @param mutableBufferManager
     */
    public UIOStreamBuilder(MutableBufferManager mutableBufferManager) {
        this.mutableBufferManager = mutableBufferManager;
        creator = new MutableBMCreator();
    }

    /**
     * always buffered
     * @param bufferSource
     */
    public UIOStreamBuilder(Buffer[] bufferSource) {
        this.bufferSource = bufferSource;
        creator = new BufferCreator();
    }

    /**
     * always buffered
     * @param url
     */
    public UIOStreamBuilder(URL url) {
        this.url = url;
        creator = new URLCreator();
    }

    /**
     * always buffered
     * @param inputStreamSource
     */
    public UIOStreamBuilder(InputStream inputStreamSource) {
        this.inputStreamSource = inputStreamSource;
        creator = new ISCreator();
    }

    /**
     * always buffered
     * @param outputStreamSource
     */
    public UIOStreamBuilder(OutputStream outputStreamSource) {
        this.outputStreamSource = outputStreamSource;
        creator = new OSCreator();
    }

    /**
     * set mode (writeable or read only)
     * @param mode READ_ONLY or READ_WRITE
     * @return this
     */
    public UIOStreamBuilder setMode(String mode) {
        if (READ_ONLY.equals(mode) || READ_WRITE.equals(mode)) {
            this.mode = mode;
            return this;
        }
        else {
            throw new IllegalArgumentException(mode);
        }
    }

    /**
     * set byte order (big endian or little endian)
     * @param byteOrder LITTLE_ENDIAN or BIG_ENDIAN (default value - BIG_ENDIAN)
     * @return this
     */
    public UIOStreamBuilder setByteOrder(int byteOrder) {
        switch (byteOrder) {
            case RandomAccess.LITTLE_ENDIAN:
            case RandomAccess.BIG_ENDIAN:
                this.byteOrder = byteOrder;
                return this;
            default:
                throw new IllegalArgumentException("" + byteOrder);
        }
    }

    /**
     * set start offset
     * @param start start offset of stream (default value - 0L)
     * @return this
     */
    public UIOStreamBuilder setStart(long start) {
        if (start < 0) {
            throw new IllegalArgumentException(" " + start);
        }
        this.start = new Long(start);
        return this;
    }

    /**
     * set stream length
     * @param length stream length
     * @return this
     */
    public UIOStreamBuilder setLength(long length) {
        if (length < 0) {
            throw new IllegalArgumentException(" " + length);
        }
        this.length = new Long(length);
        if (start == null) {
            start = new Long(0L);
        }
        return this;
    }

    /**
     * set if stream should be buffered or not (rather a hint because some streams are always buffered)
     * @param buffered true or false (default value - false)
     * @return this
     */
    public UIOStreamBuilder setBuffered(boolean buffered) {
        this.buffered = buffered;
        return this;
    }

    /**
     * finally create desired stream
     * @return RandomAccess (unbuffered, writeable), RandomAccessRO (unbuffered, read only), RandomAccessBuffer (buffered, writeable), RandomAccessBufferRO(buffered, read only) stream.
     * @throws IOException
     */
    public RandomAccessRO create() throws IOException {
        if (length != null && length.longValue() == 0) {
            Sys.err.println("Warning: stream length is 0");
        }
        if (buffered) {
            return creator.createBuffered();
        }
        else {
            return creator.create();
        }
    }

    abstract class Creator {
        abstract RandomAccessRO create() throws IOException;

        abstract RandomAccessRO createBuffered() throws IOException;

        protected int getByteOrder() {
            return byteOrder != 0 ? byteOrder : RandomAccess.BIG_ENDIAN;
        }
    }

    class FileCreator extends Creator {

        public RandomAccessRO create() throws IOException {
            if (start == null && length == null) {
                if (READ_ONLY.equals(mode)) {
                    return new RandomAccessFileWrapperRO(new RandomAccessFile(fileSource, mode), byteOrder);
                }
                else if (READ_WRITE.equals(mode)) {
                    return new RandomAccessFileWrapper(new RandomAccessFile(fileSource, mode), byteOrder);
                }
            }
            else if (length == null) {
                if (READ_ONLY.equals(mode)) {
                    return new RandomAccessFileWrapperRO(new RandomAccessFile(fileSource, mode), start.longValue(), byteOrder);
                }
                else if (READ_WRITE.equals(mode)) {
                    return new RandomAccessFileWrapper(new RandomAccessFile(fileSource, mode), start.longValue(), byteOrder);
                }
            }
            else {
                if (READ_ONLY.equals(mode)) {
                    return new RandomAccessFileWrapperRO(new RandomAccessFile(fileSource, mode), start.longValue(), length.longValue(), byteOrder);
                }
                else if (READ_WRITE.equals(mode)) {
                    return new RandomAccessFileWrapper(new RandomAccessFile(fileSource, mode), start.longValue(), length.longValue(), byteOrder);
                }
            }
            return null;
        }

        RandomAccessRO createBuffered() throws IOException {
            if (start == null && length == null) {
                if (READ_ONLY.equals(mode)) {
                    BufferManager mbm = new RAFBufferManager(new RandomAccessFile(fileSource, mode));
                    return new RandomAccessBufferRO(mbm, byteOrder);
                }
                else if (READ_WRITE.equals(mode)) {
                    MutableBufferManager mbm = new MutableRAFBufferManager(new RandomAccessFile(fileSource, mode));
                    return new RandomAccessBuffer(mbm, byteOrder);
                }
            }
            else if (length == null) {
                if (READ_ONLY.equals(mode)) {
                    BufferManager mbm = new RAFBufferManager(new OffsetRandomAccessFile(fileSource, mode, start.longValue()));
                    return new RandomAccessBufferRO(mbm, byteOrder);
                }
                else if (READ_WRITE.equals(mode)) {
                    MutableBufferManager mbm = new MutableRAFBufferManager(new OffsetRandomAccessFile(fileSource, mode, start.longValue()));
                    return new RandomAccessBuffer(mbm, byteOrder);
                }
            }
            else {
                if (READ_ONLY.equals(mode)) {
                    BufferManager mbm = new RAFBufferManager(new OffsetRandomAccessFile(fileSource, mode, start.longValue(), length.longValue()));
                    return new RandomAccessBufferRO(mbm, byteOrder);
                }
                else if (READ_WRITE.equals(mode)) {
                    MutableBufferManager mbm = new MutableRAFBufferManager(new OffsetRandomAccessFile(fileSource, mode, start.longValue(), length.longValue()));
                    return new RandomAccessBuffer(mbm, byteOrder);
                }
            }
            return null;
        }
    }

    class RAFCreator extends Creator {

        public RandomAccessRO create() throws IOException {
            if (start == null && length == null) {
                if (READ_ONLY.equals(mode)) {
                    return new RandomAccessFileWrapperRO(rafSource, getByteOrder());
                }
                else if (READ_WRITE.equals(mode)) {
                    return new RandomAccessFileWrapper(rafSource, getByteOrder());
                }
            }
            else if (length == null) {
                if (READ_ONLY.equals(mode)) {
                    return new RandomAccessFileWrapperRO(rafSource, start.longValue(), getByteOrder());
                }
                else if (READ_WRITE.equals(mode)) {
                    return new RandomAccessFileWrapper(rafSource, start.longValue(), getByteOrder());
                }
            }
            else {
                if (READ_ONLY.equals(mode)) {
                    return new RandomAccessFileWrapperRO(rafSource, start.longValue(), length.longValue(), getByteOrder());
                }
                else if (READ_WRITE.equals(mode)) {
                    return new RandomAccessFileWrapper(rafSource, start.longValue(), length.longValue(), getByteOrder());
                }
            }
            return null;
        }

        RandomAccessRO createBuffered() throws IOException {
            if (start == null && length == null) {
                if (READ_ONLY.equals(mode)) {
                    BufferManager mbm = new RAFBufferManager(new RandomAccessFile(fileSource, mode));
                    return new RandomAccessBufferRO(mbm, getByteOrder());
                }
                else if (READ_WRITE.equals(mode)) {
                    MutableBufferManager mbm = new MutableRAFBufferManager(new RandomAccessFile(fileSource, mode));
                    return new RandomAccessBuffer(mbm, getByteOrder());
                }
            }
            else if (length == null) {
                if (READ_ONLY.equals(mode)) {
                    BufferManager mbm = new RAFBufferManager(new OffsetRandomAccessFile(fileSource, mode, start.longValue()));
                    return new RandomAccessBufferRO(mbm, getByteOrder());
                }
                else if (READ_WRITE.equals(mode)) {
                    MutableBufferManager mbm = new MutableRAFBufferManager(new OffsetRandomAccessFile(fileSource, mode, start.longValue()));
                    return new RandomAccessBuffer(mbm, getByteOrder());
                }
            }
            else {
                if (READ_ONLY.equals(mode)) {
                    BufferManager mbm = new RAFBufferManager(new OffsetRandomAccessFile(fileSource, mode, start.longValue(), length.longValue()));
                    return new RandomAccessBufferRO(mbm, getByteOrder());
                }
                else if (READ_WRITE.equals(mode)) {
                    MutableBufferManager mbm = new MutableRAFBufferManager(new OffsetRandomAccessFile(fileSource, mode, start.longValue(), length.longValue()));
                    return new RandomAccessBuffer(mbm, getByteOrder());
                }
            }
            return null;
        }
    }

    class ROCreator extends Creator {

        protected int getByteOrder() {
            return byteOrder != 0 ? byteOrder : roSource.getByteOrder();
        }

        public RandomAccessRO create() throws IOException {
            if (start == null && length == null) {
                RABufferManager bm = new RABufferManager(roSource, 0L, (int) roSource.length());
                return new RandomAccessBufferRO(bm, getByteOrder());
            }
            else if (length == null) {
                RABufferManager bm = new RABufferManager(roSource, start.longValue(), (int) roSource.length());
                return new RandomAccessBufferRO(bm, getByteOrder());
            }
            else {
                RABufferManager bm = new RABufferManager(roSource, start.longValue(), length.intValue());
                return new RandomAccessBufferRO(bm, getByteOrder());
            }
        }

        RandomAccessRO createBuffered() throws IOException {
            return create();
        }
    }

    class RACreator extends Creator {

        protected int getByteOrder() {
            return byteOrder != 0 ? byteOrder : raSource.getByteOrder();
        }

        public RandomAccessRO create() throws IOException {
            if (start == null && length == null) {
                MutableRABufferManager bm = new MutableRABufferManager(raSource, 0L, (int) raSource.length());
                return new RandomAccessBuffer(bm, getByteOrder());
            }
            else if (length == null) {
                MutableRABufferManager bm = new MutableRABufferManager(raSource, start.longValue(), (int) raSource.length());
                return new RandomAccessBuffer(bm, getByteOrder());
            }
            else {
                MutableRABufferManager bm = new MutableRABufferManager(raSource, start.longValue(), length.intValue());
                return new RandomAccessBuffer(bm, getByteOrder());
            }
        }

        RandomAccessRO createBuffered() throws IOException {
            return create();
        }
    }

    class BMCreator extends Creator {
        RandomAccessRO create() throws IOException {
            return new RandomAccessBufferRO(bufferManager, getByteOrder());
        }

        RandomAccessRO createBuffered() throws IOException {
            return create();
        }
    }

    class MutableBMCreator extends Creator {
        RandomAccessRO create() throws IOException {
            return new RandomAccessBuffer(mutableBufferManager, getByteOrder());
        }

        RandomAccessRO createBuffered() throws IOException {
            return create();
        }
    }

    class BufferCreator extends Creator {
        RandomAccessRO create() throws IOException {
            if (READ_ONLY.equals(mode)) {
                return new RandomAccessBufferRO(bufferSource, getByteOrder());
            }
            else {
                return new RandomAccessBuffer(bufferSource, getByteOrder());
            }
        }

        RandomAccessRO createBuffered() throws IOException {
            return create();
        }
    }

    class ByteCreator extends Creator {
        RandomAccessRO create() throws IOException {
            if (READ_ONLY.equals(mode)) {
                if (start == null && length == null) {
                    return new RandomAccessByteArrayRO(byteSource, 0, byteSource.length, getByteOrder());
                }
                else if (length == null) {
                    return new RandomAccessByteArrayRO(byteSource, start.intValue(), byteSource.length, getByteOrder());
                }
                else {
                    return new RandomAccessByteArrayRO(byteSource, start.intValue(), length.intValue(), getByteOrder());
                }
            }
            else if (READ_WRITE.equals(mode)) {
                if (start == null && length == null) {
                    return new RandomAccessByteArray(byteSource, 0, byteSource.length, getByteOrder());
                }
                else if (length == null) {
                    return new RandomAccessByteArray(byteSource, start.intValue(), byteSource.length, getByteOrder());
                }
                else {
                    return new RandomAccessByteArray(byteSource, start.intValue(), length.intValue(), getByteOrder());
                }
            }
            return null;
        }

        RandomAccessRO createBuffered() throws IOException {
            return create();
        }
    }

    class ShortCreator extends Creator {

        RandomAccessRO create() throws IOException {
            if (READ_ONLY.equals(mode)) {
                if (start == null && length == null) {
                    return new RandomAccessBufferRO(new ShortArrayBufferManager(shortSource, 0, shortSource.length), getByteOrder());
                }
                else if (length == null) {
                    return new RandomAccessBufferRO(new ShortArrayBufferManager(shortSource, start.intValue(), shortSource.length), getByteOrder());
                }
                else {
                    return new RandomAccessBufferRO(new ShortArrayBufferManager(shortSource, start.intValue(), length.intValue()), getByteOrder());
                }
            }
            else if (READ_WRITE.equals(mode)) {
                if (start == null && length == null) {
                    return new RandomAccessBuffer(new ShortArrayBufferManager(shortSource, 0, shortSource.length), getByteOrder());
                }
                else if (length == null) {
                    return new RandomAccessBuffer(new ShortArrayBufferManager(shortSource, start.intValue(), shortSource.length), getByteOrder());
                }
                else {
                    return new RandomAccessBuffer(new ShortArrayBufferManager(shortSource, start.intValue(), length.intValue()), getByteOrder());
                }
            }
            return null;
        }

        RandomAccessRO createBuffered() throws IOException {
            return create();
        }
    }

    class CharCreator extends Creator {

        RandomAccessRO create() throws IOException {
            if (READ_ONLY.equals(mode)) {
                if (start == null && length == null) {
                    return new RandomAccessBufferRO(new CharArrayBufferManager(charSource, 0, charSource.length), getByteOrder());
                }
                else if (length == null) {
                    return new RandomAccessBufferRO(new CharArrayBufferManager(charSource, start.intValue(), charSource.length), getByteOrder());
                }
                else {
                    return new RandomAccessBufferRO(new CharArrayBufferManager(charSource, start.intValue(), length.intValue()), getByteOrder());
                }
            }
            else if (READ_WRITE.equals(mode)) {
                if (start == null && length == null) {
                    return new RandomAccessBuffer(new CharArrayBufferManager(charSource, 0, charSource.length), getByteOrder());
                }
                else if (length == null) {
                    return new RandomAccessBuffer(new CharArrayBufferManager(charSource, start.intValue(), charSource.length), getByteOrder());
                }
                else {
                    return new RandomAccessBuffer(new CharArrayBufferManager(charSource, start.intValue(), length.intValue()), getByteOrder());
                }
            }
            return null;
        }

        RandomAccessRO createBuffered() throws IOException {
            return create();
        }
    }

    class IntCreator extends Creator {

        RandomAccessRO create() throws IOException {
            if (READ_ONLY.equals(mode)) {
                if (start == null && length == null) {
                    return new RandomAccessBufferRO(new IntArrayBufferManager(intSource, 0, intSource.length), getByteOrder());
                }
                else if (length == null) {
                    return new RandomAccessBufferRO(new IntArrayBufferManager(intSource, start.intValue(), intSource.length), getByteOrder());
                }
                else {
                    return new RandomAccessBufferRO(new IntArrayBufferManager(intSource, start.intValue(), length.intValue()), getByteOrder());
                }
            }
            else if (READ_WRITE.equals(mode)) {
                if (start == null && length == null) {
                    return new RandomAccessBuffer(new IntArrayBufferManager(intSource, 0, intSource.length), getByteOrder());
                }
                else if (length == null) {
                    return new RandomAccessBuffer(new IntArrayBufferManager(intSource, start.intValue(), intSource.length), getByteOrder());
                }
                else {
                    return new RandomAccessBuffer(new IntArrayBufferManager(intSource, start.intValue(), length.intValue()), getByteOrder());
                }
            }
            return null;
        }

        RandomAccessRO createBuffered() throws IOException {
            return create();
        }
    }

    class LongCreator extends Creator {

        RandomAccessRO create() throws IOException {
            if (READ_ONLY.equals(mode)) {
                if (start == null && length == null) {
                    return new RandomAccessBufferRO(new LongArrayBufferManager(longSource, 0, longSource.length), getByteOrder());
                }
                else if (length == null) {
                    return new RandomAccessBufferRO(new LongArrayBufferManager(longSource, start.intValue(), longSource.length), getByteOrder());
                }
                else {
                    return new RandomAccessBufferRO(new LongArrayBufferManager(longSource, start.intValue(), length.intValue()), getByteOrder());
                }
            }
            else if (READ_WRITE.equals(mode)) {
                if (start == null && length == null) {
                    return new RandomAccessBuffer(new LongArrayBufferManager(longSource, 0, longSource.length), getByteOrder());
                }
                else if (length == null) {
                    return new RandomAccessBuffer(new LongArrayBufferManager(longSource, start.intValue(), longSource.length), getByteOrder());
                }
                else {
                    return new RandomAccessBuffer(new LongArrayBufferManager(longSource, start.intValue(), length.intValue()), getByteOrder());
                }
            }
            return null;
        }

        RandomAccessRO createBuffered() throws IOException {
            return create();
        }
    }

    class FloatCreator extends Creator {

        RandomAccessRO create() throws IOException {
            if (READ_ONLY.equals(mode)) {
                if (start == null && length == null) {
                    return new RandomAccessBufferRO(new FloatArrayBufferManager(floatSource, 0, floatSource.length), getByteOrder());
                }
                else if (length == null) {
                    return new RandomAccessBufferRO(new FloatArrayBufferManager(floatSource, start.intValue(), floatSource.length), getByteOrder());
                }
                else {
                    return new RandomAccessBufferRO(new FloatArrayBufferManager(floatSource, start.intValue(), length.intValue()), getByteOrder());
                }
            }
            else if (READ_WRITE.equals(mode)) {
                if (start == null && length == null) {
                    return new RandomAccessBuffer(new FloatArrayBufferManager(floatSource, 0, floatSource.length), getByteOrder());
                }
                else if (length == null) {
                    return new RandomAccessBuffer(new FloatArrayBufferManager(floatSource, start.intValue(), floatSource.length), getByteOrder());
                }
                else {
                    return new RandomAccessBuffer(new FloatArrayBufferManager(floatSource, start.intValue(), length.intValue()), getByteOrder());
                }
            }
            return null;
        }

        RandomAccessRO createBuffered() throws IOException {
            return create();
        }
    }

    class DoubleCreator extends Creator {

        RandomAccessRO create() throws IOException {
            if (READ_ONLY.equals(mode)) {
                if (start == null && length == null) {
                    return new RandomAccessBufferRO(new DoubleArrayBufferManager(doubleSource, 0, doubleSource.length), getByteOrder());
                }
                else if (length == null) {
                    return new RandomAccessBufferRO(new DoubleArrayBufferManager(doubleSource, start.intValue(), doubleSource.length), getByteOrder());
                }
                else {
                    return new RandomAccessBufferRO(new DoubleArrayBufferManager(doubleSource, start.intValue(), length.intValue()), getByteOrder());
                }
            }
            else if (READ_WRITE.equals(mode)) {
                if (start == null && length == null) {
                    return new RandomAccessBuffer(new DoubleArrayBufferManager(doubleSource, 0, doubleSource.length), getByteOrder());
                }
                else if (length == null) {
                    return new RandomAccessBuffer(new DoubleArrayBufferManager(doubleSource, start.intValue(), doubleSource.length), getByteOrder());
                }
                else {
                    return new RandomAccessBuffer(new DoubleArrayBufferManager(doubleSource, start.intValue(), length.intValue()), getByteOrder());
                }
            }
            return null;
        }

        RandomAccessRO createBuffered() throws IOException {
            return create();
        }
    }

    class URLCreator extends Creator {

        RandomAccessRO create() throws IOException {
            return new RandomAccessBufferRO(new HTTPBufferManager(url), getByteOrder());
        }

        RandomAccessRO createBuffered() throws IOException {
            return create();
        }
    }

    class ISCreator extends Creator {
        RandomAccessRO create() throws IOException {
            BufferManager bm = new InputStreamBufferManager(inputStreamSource);
            return new RandomAccessBufferRO(bm, getByteOrder());
        }

        RandomAccessRO createBuffered() throws IOException {
            return create();
        }
    }

    class OSCreator extends Creator {
        RandomAccessRO create() throws IOException {
            MutableBufferManager bm = new OutputStreamBufferManager(outputStreamSource);
            return new RandomAccessBuffer(bm, getByteOrder());
        }

        RandomAccessRO createBuffered() throws IOException {
            return create();
        }
    }
}
